/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
package org.processmining.xeslite.external;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.XVisitor;

import com.google.common.collect.ForwardingList;

/**
 * @author F. Mannhardt
 *
 */
final class XLogExternalImpl extends ForwardingList<XTrace> implements XLog, ExternalAttributable {
	
	private final long id;
	private ExternalStore store;
	
	private List<XTrace> traces;
	
	private Set<XExtension> extensions;
	private List<XEventClassifier> classifiers;
	private List<XAttribute> globalTraceAttributes;
	private List<XAttribute> globalEventAttributes;
	private XEventClassifier cachedClassifier;
	private XLogInfo cachedInfo;

	XLogExternalImpl(ExternalStore attributeStore) {
		this(null, attributeStore);
	}
	
	XLogExternalImpl(XAttributeMap attributeMap, ExternalStore attributeStore) {
		this(attributeStore.getIdFactory().nextId(), attributeMap, attributeStore);
	}
	
	XLogExternalImpl(long id, XAttributeMap attributeMap, ExternalStore attributeStore) {
		this.store = attributeStore;
		this.id = id;
		this.traces = new ArrayList<>();
		if (attributeMap != null) {
			XAttributeMap newAttributeMap = getAttributes();
			for (XAttribute a: attributeMap.values()) {
				newAttributeMap.put(a.getKey(), a);
			}
		}
		
		// Transient fields
		this.extensions = new HashSet<>();
		this.classifiers = new ArrayList<>();
		this.globalTraceAttributes = new ArrayList<>();
		this.globalEventAttributes = new ArrayList<>();
		this.cachedClassifier = null;
		this.cachedInfo = null;
	}
	
	
	public long getExternalId() {
		return id;
	}

	@Override
	public XAttributeMap getAttributes() {
		return store.getAttributes(this);
	}

	@Override
	public void setAttributes(XAttributeMap attributes) {
		store.setAttributes(this, attributes);
	}
	
	@Override
	public boolean hasAttributes() {
		return store.hasAttributes(this);
	}
	
	/* (non-Javadoc)
	 * @see org.deckfour.xes.model.XAttributable#getExtensions()
	 */
	public Set<XExtension> getExtensions() {
		return extensions;
	}
	
	/* (non-Javadoc)
	 * @see org.deckfour.xes.model.XLog#getClassifiers()
	 */
	public List<XEventClassifier> getClassifiers() {
		return classifiers;
	}

	/* (non-Javadoc)
	 * @see org.deckfour.xes.model.XLog#getGlobalEventAttributes()
	 */
	public List<XAttribute> getGlobalEventAttributes() {
		return globalEventAttributes;
	}

	/* (non-Javadoc)
	 * @see org.deckfour.xes.model.XLog#getGlobalTraceAttributes()
	 */
	public List<XAttribute> getGlobalTraceAttributes() {
		return globalTraceAttributes;
	}

	/* (non-Javadoc)
	 * @see org.deckfour.xes.model.XLog#getInfo(org.deckfour.xes.classification.XEventClassifier)
	 */
	public XLogInfo getInfo(XEventClassifier classifier) {
		return classifier.equals(cachedClassifier) ? cachedInfo : null;
	}

	/* (non-Javadoc)
	 * @see org.deckfour.xes.model.XLog#setInfo(org.deckfour.xes.classification.XEventClassifier, org.deckfour.xes.info.XLogInfo)
	 */
	public void setInfo(XEventClassifier classifier, XLogInfo info) {
		cachedClassifier = classifier;
		cachedInfo = info;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	@Override
	public Object clone() {
		XLogExternalImpl clone = new XLogExternalImpl(getAttributes(), store);
		clone.extensions = new HashSet<>(extensions);
		clone.classifiers = new ArrayList<>(classifiers);
		clone.globalTraceAttributes = new ArrayList<>(globalTraceAttributes);
		clone.globalEventAttributes = new ArrayList<>(globalEventAttributes);
		clone.cachedClassifier = null;
		clone.cachedInfo = null;
		for (XTrace trace : this) {
			clone.add((XTrace) trace.clone());
		}
		return clone;
	}
	
	/*
	 * Runs the given visitor on this log.
	 * 
	 * (non-Javadoc)
	 * 
	 * @see org.deckfour.xes.model.XLog#accept(org.deckfour.xes.model.XVisitor)
	 */
	public boolean accept(XVisitor visitor) {
		/*
		 * Check whether the visitor may run.
		 */
		if (visitor.precondition()) {
			/*
			 * Yes, it may. Now initialize.
			 */
			visitor.init(this);
			/*
			 * First call.
			 */
			visitor.visitLogPre(this);
			/*
			 * Visit the extensions.
			 */
			for (XExtension extension : extensions) {
				extension.accept(visitor, this);
			}
			/*
			 * Visit the classifiers.
			 */
			for (XEventClassifier classifier : classifiers) {
				classifier.accept(visitor, this);
			}
			/*
			 * Visit the attributes.
			 */
			if (hasAttributes()) {
				for (XAttribute attribute : getAttributes().values()) {
					attribute.accept(visitor, this);
				}
			}
			/*
			 * Visit the traces.
			 */
			for (XTrace trace : this) {
				trace.accept(visitor, this);
			}
			/*
			 * Last call.
			 */
			visitor.visitLogPost(this);
			return true;
		}
		return false;
	}

	protected List<XTrace> delegate() {
		return traces;
	}

}
