package org.processmining.xeslite.query;

import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.XVisitor;
import org.deckfour.xes.util.XAttributeUtils;
import org.processmining.log.utils.XUtils;
import org.processmining.xeslite.query.attributes.DateAttributeIndex;
import org.processmining.xeslite.query.attributes.DoubleAttributeIndex;
import org.processmining.xeslite.query.attributes.LiteralAttributeIndex;
import org.processmining.xeslite.query.syntax.ParseException;
import org.processmining.xeslite.query.syntax.QueryParser;
import org.processmining.xeslite.query.syntax.QueryParserVisitor;
import org.processmining.xeslite.query.syntax.QueryRoot;

import com.google.common.base.Function;
import com.google.common.collect.Collections2;
import com.google.common.collect.ForwardingList;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Iterables;
import com.google.common.collect.Iterators;
import com.googlecode.cqengine.ConcurrentIndexedCollection;
import com.googlecode.cqengine.IndexedCollection;
import com.googlecode.cqengine.attribute.Attribute;
import com.googlecode.cqengine.attribute.MultiValueNullableAttribute;
import com.googlecode.cqengine.attribute.SimpleAttribute;
import com.googlecode.cqengine.index.Index;
import com.googlecode.cqengine.index.radixinverted.InvertedRadixTreeIndex;
import com.googlecode.cqengine.query.Query;
import com.googlecode.cqengine.query.QueryFactory;
import com.googlecode.cqengine.query.option.QueryOptions;
import com.googlecode.cqengine.resultset.ResultSet;
import com.googlecode.cqengine.resultset.common.WrappedResultSet;

class XIndexedTracesImpl implements XIndexedTraces {

	private static final class XUnwrappingResultSet extends WrappedResultSet<XTrace> {

		public XUnwrappingResultSet(ResultSet<XTrace> wrappedResultSet, QueryOptions queryOptions) {
			super(wrappedResultSet, queryOptions);
		}

		public Iterator<XTrace> iterator() {
			return Iterators.transform(super.iterator(), new Function<XTrace, XTrace>() {

				public XTrace apply(XTrace trace) {
					return unwrapTrace((XWrappedTrace) trace);
				}
			});
		}

		public boolean contains(XTrace object) {
			return super.contains(wrapTrace(object));
		}

	}

	private static final class XWrappedTrace extends ForwardingList<XEvent> implements XTrace {

		private final XTrace trace;
		private final String name;

		public XWrappedTrace(XTrace trace) {
			this.trace = trace;
			this.name = XConceptExtension.instance().extractName(trace);
			if (name == null) {
				throw new IllegalArgumentException("Trace needs a 'concept:name' to be included in XIndexedTraces.");
			}
		}

		public XTrace getWrappedTrace() {
			return trace;
		}

		@Override
		public XAttributeMap getAttributes() {
			return trace.getAttributes();
		}

		@Override
		public Set<XExtension> getExtensions() {
			return trace.getExtensions();
		}

		@Override
		public boolean hasAttributes() {
			return trace.hasAttributes();
		}

		@Override
		public void setAttributes(XAttributeMap arg0) {
			trace.setAttributes(arg0);
		}

		@Override
		public void accept(XVisitor arg0, XLog arg1) {
			trace.accept(arg0, arg1);
		}

		@Override
		public int insertOrdered(XEvent arg0) {
			return trace.insertOrdered(arg0);
		}

		@Override
		public Object clone() {
			return trace.clone();
		}

		@Override
		protected List<XEvent> delegate() {
			return trace;
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((name == null) ? 0 : name.hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (!(obj instanceof XWrappedTrace))
				return false;
			XWrappedTrace other = (XWrappedTrace) obj;
			if (name == null) {
				if (other.name != null)
					return false;
			} else if (!name.equals(other.name))
				return false;
			return true;
		}

	}

	private static XTrace wrapTrace(XTrace element) {
		return new XWrappedTrace(element);
	}

	private static XTrace unwrapTrace(XWrappedTrace trace) {
		return trace.getWrappedTrace();
	}

	private final IndexedCollection<XTrace> indexedCollection;

	XIndexedTracesImpl(Collection<XTrace> log) {
		indexedCollection = new ConcurrentIndexedCollection<>();
		indexedCollection.addIndex(InvertedRadixTreeIndex.onAttribute(eventNames()));
		// Bulk-add
		indexedCollection.addAll(Collections2.transform(log, new Function<XTrace, XTrace>() {

			public XTrace apply(XTrace trace) {
				return wrapTrace(trace);
			}
		}));
	}

	public void addIndex(Index<XTrace> index) {
		indexedCollection.addIndex(index);
	}

	public ResultSet<XTrace> retrieve(Query<XTrace> query) {
		return unwrapResult(indexedCollection.retrieve(query), QueryFactory.noQueryOptions());
	}

	@SuppressWarnings("unchecked")
	public Query<XTrace> parseQuery(String queryString, AttributeTypeResolver attributeTypeResolver)
			throws ParseException {
		if (!queryString.isEmpty() && queryString.charAt(0) == '~') {
			if (queryString.length() > 1) {
				queryString = queryString.substring(1);
				return QueryFactory.or(QueryFactory.matchesRegex(traceName(), queryString),
						QueryFactory.matchesRegex(eventNames(), queryString));
			} else {
				throw new ParseException("Missing regular expression afterr '~'");
			}
		} else if (!queryString.isEmpty()) {
			QueryParser queryParser = new QueryParser(queryString);
			try {
				QueryRoot rootNode = queryParser.parse();
				QueryParserVisitor visitor = new QueryBuilderVisitor<XTrace>(this, attributeTypeResolver);
				return (Query<XTrace>) rootNode.jjtAccept(visitor, null);
			} catch (ParseException e) {
				Pattern queryPattern = Pattern.compile("(=|>|<|~|<=|>=|\\!=|and|or)", Pattern.CASE_INSENSITIVE);
				if (!queryPattern.matcher(queryString).find()) {
					// Simple query without any operator -> search in names
					return QueryFactory.or(QueryFactory.equal(traceName(), queryString),
							QueryFactory.equal(eventNames(), queryString));
				} else {
					throw e;
				}
			}
		} else {
			return QueryFactory.all(XTrace.class);
		}
	}

	public Query<XTrace> parseQuery(String queryString) throws ParseException {
		return parseQuery(queryString, new AttributeTypeResolver() {

			public Class<? extends XAttribute> getAttributeType(final String attributeName) {
				// simply return the first class found
				for (XTrace t : indexedCollection) {
					XAttribute attribute = t.getAttributes().get(attributeName);
					if (attribute != null) {
						return XAttributeUtils.getType(attribute);
					}
					for (XEvent e : t) {
						attribute = e.getAttributes().get(attributeName);
						if (attribute != null) {
							return XAttributeUtils.getType(attribute);
						}
					}
				}
				return XAttributeLiteral.class;

			}
		});
	}

	public ResultSet<XTrace> retrieve(String queryString) throws ParseException {
		return unwrapResult(indexedCollection.retrieve(parseQuery(queryString)), QueryFactory.noQueryOptions());
	}

	private ResultSet<XTrace> unwrapResult(ResultSet<XTrace> result, QueryOptions queryOptions) {
		return new XUnwrappingResultSet(result, queryOptions);
	}

	public boolean remove(XTrace element) {
		return indexedCollection.remove(wrapTrace(element));
	}

	public void add(XTrace element) {
		indexedCollection.add(wrapTrace(element));
	}

	/**
	 * @return Attribute concept:name on the trace
	 */
	public final SimpleAttribute<XTrace, String> traceName() {
		return new SimpleAttribute<XTrace, String>("trace.concept:name") {

			public String getValue(XTrace trace, QueryOptions queryOptions) {
				return XUtils.getConceptName(trace);
			}
		};
	}

	/**
	 * @return Attribute concept:name of events contained in traces
	 */
	public final Attribute<XTrace, String> eventNames() {
		return new MultiValueNullableAttribute<XTrace, String>("event.concept:name", true) {

			public Iterable<String> getNullableValues(XTrace trace, QueryOptions queryOptions) {
				return Iterables.transform(trace, new Function<XEvent, String>() {

					public String apply(XEvent event) {
						return XUtils.getConceptName(event);
					}
				});
			}

		};
	}

	/**
	 * String attribute for {@link XAttributeLiteral} and
	 * {@link XAttributeBoolean}
	 * 
	 * @param attributeName
	 * @return
	 */
	public Attribute<XTrace, String> stringAttribute(String attributeName) {
		return new LiteralAttributeIndex<XTrace>(XTrace.class, attributeName) {

			protected Iterable<XAttributable> getAttributables(XTrace trace) {
				return extractAttributables(trace);
			}

		};
	}

	/**
	 * Double attribute for {@link XAttributeContinuous} and
	 * {@link XAttributeDiscrete}
	 * 
	 * @param attributeName
	 * @return
	 */
	public Attribute<XTrace, Double> doubleAttribute(String attributeName) {
		return new DoubleAttributeIndex<XTrace>(XTrace.class, attributeName) {

			protected Iterable<XAttributable> getAttributables(XTrace trace) {
				return extractAttributables(trace);
			}

		};
	}

	/**
	 * Date attribute
	 * 
	 * @param attributeName
	 * @return
	 */
	public Attribute<XTrace, Date> dateAttribute(String attributeName) {
		return new DateAttributeIndex<XTrace>(XTrace.class, attributeName) {

			protected Iterable<XAttributable> getAttributables(XTrace trace) {
				return extractAttributables(trace);
			}

		};
	}

	private Iterable<XAttributable> extractAttributables(XTrace trace) {
		Iterable<XAttributable> events = Iterables.transform(trace, new Function<XEvent, XAttributable>() {

			public XAttributable apply(XEvent event) {
				return event;
			}
		});
		return Iterables.concat(ImmutableList.of(trace), events);
	}
}
